##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
'''Unified SSH client for IBM Spectrum Virtualize Family Storage'''

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import pkg_resources
from logging import getLogger
import pysvc.errors as ce
from pysvc.messages import UnifiedMessages
from pysvc.transports.ssh_transport import SSHTransport
from pysvc.unified.clispec import parse
from pysvc import PYSVC_DEFAULT_LOGGER
from .scp_cli_client import ScpClient
from pysvc.unified.helpers import etree
from pysvc.unified.helpers.xml_util import XMLException

__all__ = ['connect']

xlog = getLogger(PYSVC_DEFAULT_LOGGER)


class IncorrectDeviceTypeError(ce.UnableToConnectException):
    '''Raise if the expected device type is not found.'''
    pass


class NoSpecificationError(ce.UnableToConnectException):
    '''Raise if no CLI specification is found for the storage array.'''
    pass


class UnifiedSSHClient(object):
    '''Unified SSH client for IBM Spectrum Virtualize Family Storage.

    It creates callable stubs for the CLI commands of storage array.
    The stub is implemented as :py:class:`.Proxy` for
    :py:class:`pysvc.pysvc.unified.clispec.CLICommand`.
    '''

    def __init__(self):
        super(UnifiedSSHClient, self).__init__()
        self.transport = None
        self.specification = None
        self.flexible = False

    def close(self):
        '''Close the connection.'''
        if self.transport:
            self.transport.disconnect()
            self.transport = None
        self.specification = None

    def send_raw_command(self, cmd, extra=None, stdin=None):
        '''Send plain string as command to storage array and execute.

        :param cmd: The command.
        :type cmd: str
        :param extra: (optional) The extra parameters.

                      * timeout: (float) Response timeout for each sending
                        command in seconds.
        :type extra: dict
        :return: The content from stdout and stderr of the executed command.
        :rtype: tuple
        '''
        timeout = extra.get('timeout', 0) if extra else 0
        xlog.debug("+++{0}+++".format(cmd))
        _, stdout, stderr = self.transport.send_command(
            cmd, raw=True, timeout=timeout, stdin_input=stdin)
        return stdout, stderr

    def get_device_info(self):
        '''Get the device information of storage array.

        :return: The device type and version.
        :rtype: tuple
        '''
        return (self.specification.array_type,
                self.specification.array_infos) if self.specification else None

    def get_dump_element_tree(self, remote_path, timeout=None):
        """
        get the element tree of xml remote_path specified
        :param remote_path: full path on the SVC,
               e.g. /dumps/iostats/Nn_stats_151240_151120_162817
        :param timeout: int, the timeout of the session
        :return: ElementTree
        """
        raw_xml = self.get_dump(remote_path, timeout)
        try:
            return etree.parse(StringIO(raw_xml))
        except XMLException as ex:
            err_msg = (r'The dump file context is not valid XML: {}'
                       .format(raw_xml))
            xlog.error(err_msg)
            raise ex(err_msg)

    def get_dump(self, remote_path, timeout=None):
        """
        :param remote_path: full path on the SVC,
               e.g. /dumps/iostats/Nn_stats_151240_151120_162817
        :param timeout: int, the timeout of the session
        :return: str, the context of the file
        """
        scp_client = ScpClient(self.transport.transport.get_transport(),
                               timeout)
        return scp_client.receive(remote_path)

    def __getattr__(self, name):
        obj = getattr(self.specification, name, None)
        if obj is None:
            raise AttributeError(
                "'%s' object has no attribute '%s'" %
                (self.__class__.__name__, name))
        return Proxy(obj, self.send_raw_command)

    def __dir__(self):
        return dir(self.specification)


class Proxy(object):
    '''Proxy for CLI command

    Read __doc__ attribute to get document instead of
    using __builtins__.help(), e.g.

    >>> print a_proxy.__doc__
    Help on ...
    '''

    def __init__(self, referent, context):
        super(Proxy, self).__init__()
        self.referent = referent
        self.context = context

    @property
    def __doc__(self):
        return getattr(self.referent, '__doc__', None)

    def __getattr__(self, name):
        at = getattr(self.referent, name, None)
        if at is None:
            raise AttributeError(
                "'%s' object has no attribute '%s'" %
                (self.__class__.__name__, name))
        return Proxy(at, self.context)

    def __dir__(self):
        return dir(self.referent)

    def __call__(self, **kwargs):
        '''Call the wrapped referent with given parameters and
           return the result.'''
        return self.referent(self.context, kwargs)


def yield_device_type(conn):
    conn.specification, oldspec = parse_cli_spec(
        conn, get_cli_spec('xsf', '1.0')), conn.specification
    try:
        try:
            for clu in conn.cli.lscluster():
                device = 'ifs' if clu.get('Profile') == 'IFS' else 'sonas'
                for nd in conn.cli.lsnode(cluster=clu.get('Name')):
                    yield device, canonical_version(
                        nd.get('Product Version', '')
                        or nd.get('Product version', '')
                        or nd.get('product version', ''))
        except GeneratorExit:
            raise
        except Exception:
            xlog.exception('No IFS or SoNAS is found, and continue.')
        try:
            for clu in conn.svcinfo.lscluster():
                if clu.get('location') == 'local':
                    for clu1 in conn.svcinfo.lscluster(cluster=clu.get('id')):
                        yield 'svc', canonical_version(
                            clu1.get('code_level', ''))
        except GeneratorExit:
            raise
        except Exception:
            xlog.exception('No SVC or Storwize is found, and continue.')
    finally:
        conn.specification = oldspec


def set_specification(conn, with_remote_clispec=True):
    spec = get_remote_cli_spec(conn) if with_remote_clispec else None
    if not spec:
        xlog.info(UnifiedMessages.UNIFIED_PARSE_LOCAL_START)
        for d, t in yield_device_type(conn):
            try:
                spec = parse_cli_spec(conn, get_cli_spec(d, t))
            except Exception:
                xlog.exception(
                    'No CLI specification found for "%s, %s", and continue.' %
                    (d, t))
            if spec:
                break
    if not spec:
        raise NoSpecificationError(UnifiedMessages.UNIFIED_NO_CLI_SPEC)
    conn.specification = spec


def check_device_type(conn, device_type):
    if not device_type:
        return
    info = conn.get_device_info()
    if not info or info[0] != device_type:
        raise IncorrectDeviceTypeError(
            UnifiedMessages.UNIFIED_INCORRECT_DEVICE_TYPE(device_type))


def get_remote_cli_spec(conn):
    try:
        stdout, stderr = conn.send_raw_command('catxmlspec')
        if stdout:
            return parse_cli_spec(conn, StringIO(stdout.decode()))
        xlog.warning(UnifiedMessages.UNIFIED_CATXMLSPEC_FAIL(stderr))
    except Exception:
        xlog.exception(UnifiedMessages.UNIFIED_PARSE_REMOTE_FAIL)


def parse_cli_spec(conn, source):
    spec = parse(source, flexible=conn.flexible)
    # make sure there is CLI command defined in CLI spec
    return spec if spec and spec.cmds else None


def device_type_alias(device, version):
    if device in ('storwize', 'storwise'):
        device = 'svc'
    if device == 'svc':
        if version.startswith('6.') and version not in ('6.1', '6.2', '6.3'):
            version = '6.3'
    return device, version


def get_cli_spec(device, version):
    # pylint: disable-msg=E1101
    return pkg_resources.resource_stream(
        __name__, '%s-%s.xml' %
        device_type_alias(
            device, version))


def canonical_version(data):
    return '.'.join(data.strip().split('.')[:2])


def connect(address, **kwargs):
    '''Connect to storage array through SSH.

    :param address: The IP address or host name of storage array.
    :type address: str
    :param username: (optional) The username of login account.
    :type username: str
    :param password: (optional) The password of login account or private key.
    :type password: str
    :param port: (optional) The port of SSH service in storage array,
                            it is 22 by default.
    :type port: int
    :param privatekey: (optional) The private key of login account.
    :type privatekey: str
    :param privatekey_filename: (optional) The file name of private key.
    :type privatekey_filename: str
    :param add_hostkey: (optional) Indicates whether to add SSH host key of
           storage array to known hosts, it is True by default.
    :type add_hostkey: bool
    :param timeout: (optional) Connection timeout in seconds,
                               it is 30 by default.
    :type timeout: int
    :param cmd_timeout: (optional) Response timeout for each sending command
                                   in seconds, it is 60.0 by default.
    :type cmd_timeout: float
    :param device_type: (optional) The device type of storage array, e.g.
                                   "svc", "storwize", None.
    :type device_type: str
    :param flexible: (optional) Indicates whether to enable flexible mode
                                which bypasses strict error checking, it is
                                False by default.
    :type flexible: bool
    :param with_remote_clispec: (optional) Indicates whether to read CLI
                                           specification from remote storage
                                           array, it is True by default.
    :type with_remote_clispec: bool
    :return: The connection object to storage array.
    :rtype: :py:class:`.UnifiedSSHClient`

    Example:

    >>> connect('ip', username='admin', privatekey_filename=r'/local/key')
    <pysvc.pysvc.unified.client.UnifiedSSHClient object at 0x...>
    '''
    g = kwargs.get
    trans = SSHTransport(
        host=address,
        user=g('username'),
        password=g('password'),
        port=g(
            'port',
            22),
        auto_add=g(
            'add_hostkey',
            True),
        pkey=g('privatekey'),
        pkey_file=g('privatekey_filename'),
        timeout=g(
            'timeout',
            30),
        cmd_timeout=g(
            'cmd_timeout',
            60.0))
    conn = UnifiedSSHClient()
    try:
        trans.connect()
        conn.flexible = g('flexible', False)
        conn.transport = trans
        set_specification(conn, g('with_remote_clispec', True))
        check_device_type(conn, g('device_type'))
        return conn
    except BaseException:
        trans.disconnect()
        conn.close()
        raise



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from paramiko.ssh_exception import SSHException
from pysvc.transports.transport import CommonTransport
from paramiko import PKey
from paramiko import SSHClient
from paramiko.ssh_exception import PasswordRequiredException
from paramiko.ssh_exception import PartialAuthentication
from paramiko.ssh_exception import BadHostKeyException
from paramiko.ssh_exception import BadAuthenticationType
from paramiko.ssh_exception import AuthenticationException
from pysvc.errors import ConnectionTimedoutException
from pysvc.errors import TransportMessages
from pysvc.errors import IncorrectCredentials
from pysvc.errors import BadAuthenticationTypeException
from pysvc.errors import BadHostFingerPrintException
from pysvc.errors import PartialAuthenticationException
from pysvc.errors import PassphraseRequiredException
from pysvc.errors import HostDoesNotExistException
from pysvc.errors import UnableToConnectException
from logging import getLogger
from contextlib import contextmanager
from pysvc import PYSVC_DEFAULT_LOGGER
import paramiko
import socket
import os
import warnings
warnings.filterwarnings('ignore', category=DeprecationWarning)


xlog = getLogger(PYSVC_DEFAULT_LOGGER)


class SSHTransport(CommonTransport):
    DEFAULT_SSH_PORT = 22

    def __init__(
            self,
            host,
            user=None,
            password=None,
            pkey=None,
            pkey_file=None,
            port=22,
            timeout=30,
            auto_add=True,
            cmd_timeout=30.0,
            ignore_known_hosts=True):
        """
        Constructor for common SSH trasport class.
        pkey        : Key object used to sign and verify SSH2 data;
        pkey_file   : File name of the private keys to try for
                      SSH authentication;
        timeout     : Timeout value for establishing a ssh connection;
        auto_add    : If the specified mamagement node cannot be identified,
                      whether I should auto add it to known host list and
                      save to a local file;
        cmd_timeout : Timeout valeu for a command to return;
        """
        super(SSHTransport, self).__init__()
        self.host = host
        self.user = user
        self.password = password
        self.pkey = pkey
        self.private_key = pkey_file
        self.port = port
        self.timeout = timeout
        self.cmd_exec_timeout = cmd_timeout
        self.connected_endpoint = None
        self.auto_add_unknown_hosts = auto_add
        self.transport = SSHClient()
        self.sftp_client = None
        self.is_client_connected = False
        self.is_known_hosts_ignored = ignore_known_hosts
        self.svc_client_host_keys_file = \
            "%s/xsf_known_hosts" % os.path.expanduser("~")
        if self.pkey is not None:
            if not isinstance(self.pkey, PKey):
                xlog.debug(TransportMessages.SSH_INCORRECT_PRIVATE_KEY)
                raise IncorrectCredentials(
                    message=TransportMessages.SSH_INCORRECT_PRIVATE_KEY)

        if (os.path.isfile(self.svc_client_host_keys_file) and
                self.is_known_hosts_ignored is False):
            try:
                self.transport.load_host_keys(
                    filename=self.svc_client_host_keys_file)
            except IOError as ex:
                xlog.debug(ex)
                raise IncorrectCredentials(
                    message=TransportMessages.SSH_FAILED_TO_LOAD_KNOWN_HOSTS,
                    original_exception=ex)

    def __str__(self):
        if self.port == self.DEFAULT_SSH_PORT:
            return self.host
        return 'ssh://%s:%s' % (self.host, self.port)

    def __repr__(self):
        return '<%s>' % self

    @contextmanager
    def _exception_handler(self):
        try:
            yield
        except BadAuthenticationType as ex:
            xlog.debug(ex)
            raise BadAuthenticationTypeException(
                allowed_types=ex.allowed_types, original_exception=ex)
        except BadHostKeyException as ex:
            xlog.debug(ex)
            raise BadHostFingerPrintException(
                hostname=ex.hostname,
                expected_key=ex.expected_key,
                presented_key=ex.key,
                original_exception=ex)
        except PartialAuthentication as ex:
            xlog.debug(ex)
            raise PartialAuthenticationException(
                allowed_types=ex.allowed_types, original_exception=ex)
        except PasswordRequiredException as ex:
            xlog.debug(ex)
            raise PassphraseRequiredException(
                message=TransportMessages.SSH_PASS_PHRASE_REQUIRED,
                original_exception=ex)
        except AuthenticationException as ex:
            xlog.debug(ex)
            raise IncorrectCredentials(
                message=TransportMessages.SSH_AUTHENTICATION_FAILURE,
                original_exception=ex)
        except socket.timeout as ex:
            xlog.debug(ex)
            raise ConnectionTimedoutException(
                message=TransportMessages.SSH_CONNECT_TIMED_OUT,
                original_exception=ex)
        except (socket.herror, socket.gaierror) as ex:
            xlog.debug(ex)
            raise HostDoesNotExistException(
                hostname=self.host, original_exception=ex)
        except (SSHException, socket.error) as ex:
            xlog.debug(ex)
            raise UnableToConnectException(
                message=TransportMessages.SSH_UNABLE_TO_CONNECT(
                    self.host), original_exception=ex)

    def connect(self):
        """
        Initialize a SSH connection with properties
        which were set up in constructor.
        """
        with self._exception_handler():
            if self.auto_add_unknown_hosts:
                self.transport.set_missing_host_key_policy(
                    paramiko.AutoAddPolicy())
            self.transport.connect(
                self.host,
                port=self.port,
                username=self.user,
                password=self.password,
                pkey=self.pkey,
                key_filename=self.private_key,
                timeout=self.timeout)
            self.connected_endpoint = self.host
            self.is_client_connected = True
            if self.is_known_hosts_ignored is False:
                self.transport.save_host_keys(self.svc_client_host_keys_file)

    def is_connected(self):
        return self.is_client_connected

    def disconnect(self):
        """
        Disconnect from the SSH server.
        """
        if self.is_client_connected:
            self.transport.close()
            self.is_client_connected = False

    def reconnect(self):
        self.disconnect()
        self.connect()

    def send_command(
            self,
            command,
            buf_size=-1,
            raw=False,
            timeout=0,
            stdin_input=None):
        with self._exception_handler():
            # return self.transport.exec_command(command)
            try:
                channel = self.transport.get_transport().open_session()
                channel.settimeout(timeout or self.cmd_exec_timeout)
                channel.exec_command(command)
                stdin = channel.makefile('wb', buf_size)
                stdout = channel.makefile('rb', buf_size)
                stderr = channel.makefile_stderr('rb', buf_size)
                if stdin_input is not None:
                    stdin.write(stdin_input)
                    stdin.flush()
                    # shutdown_write to close write channel, paramiko will send
                    # EOF to device.
                    channel.shutdown_write()
                if raw:  # gain performance without spliting line
                    return stdin, stdout.read(), stderr.read()
                return stdin, stdout.readlines(), stderr.readlines()
            except socket.timeout as ex:
                # Need to reconnect to terminate the remote method call
                self.reconnect()
                xlog.error(ex)
                raise ConnectionTimedoutException(
                    message=TransportMessages.SSH_CON_TIMED_OUT_WHEN_EXEC_CMD,
                    original_exception=ex)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
'''Parsers for CLI response'''

import csv
from functools import reduce
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    long
except NameError:
    long = int

try:
    basestring
except NameError:
    basestring = str

from munch import Munch
from logging import getLogger
import pysvc.errors as ce
from pysvc import PYSVC_DEFAULT_LOGGER

__all__ = ['find_response_helper']

DEFAULT_DELIM = '\t'
xlog = getLogger(PYSVC_DEFAULT_LOGGER)


class CLIFailureError(ce.StorageArrayClientException):
    '''Raise if CLI fails.'''

    def __init__(self, message=None, original_exception=None, returnCode=-1):
        ce.StorageArrayClientException.__init__(
            self, message, original_exception=None)
        self.returnCode = returnCode


class CLIResponse(object):
    '''Iterable for CLI response.

    Example:

    >>> resp
    <pysvc.pysvc.unified.response.CLIResponse object at 0x...>
    >>> for part in resp:
    ...     print part
    ...
    Bunch(..., name = 'a', ...)
    Bunch(..., name = 'b', ...)
    >>>
    >>> list(resp) # or resp.as_list
    [Bunch(..., name = 'a', ...), Bunch(..., name = 'b', ...)]
    >>>
    >>> resp.as_single_element
    Bunch(..., name = ['a', 'b'], ...)
    >>>
    >>> resp.as_dict('name')
    Bunch(a = Bunch(..., name = 'a', ...), b = Bunch(..., name = 'b', ...))
    '''

    def __init__(self, resp, kwargs=None):
        super(CLIResponse, self).__init__()
        if kwargs is None:
            kwargs = {}
        self.response = resp
        try:
            self.result = self.parse(resp, kwargs)
        except Exception:
            if kwargs.get('flexible', False):
                self.result = tuple()
                xlog.exception(
                    'Fail to parse CLI output, but continue in flexible mode.')
            else:
                raise

    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        stdout = stdout.lstrip()  # remove starting space to work with csv
        stdoutlines = stdout.splitlines()

        # e.g. if stdout is "<error_tag> 4", the return code is 4
        error_tag = kwargs.get('error_tag', '')
        if error_tag:
            for line in stdoutlines:
                idx = line.find(error_tag)
                if idx >= 0:
                    rc = ' '.join(
                        tk for tk in line[idx + len(error_tag):].split()
                        if tk.isdigit() or tk.startswith('-')
                        and tk[1:].isdigit())
                    try:
                        returnCode = int(rc)
                    except Exception as e:
                        xlog.error(e)
                        returnCode = -1
                    raise CLIFailureError(
                        'CLI failure. Return code is %s. '
                        'Error message is "%s"' % (rc, stderr),
                        returnCode=returnCode)

        delim = kwargs.get('delim', DEFAULT_DELIM)

        result = []
        snif = MySniffer(delim)
        rd = csv.reader(stdoutlines, snif.sniff(stdout))
        with_header = kwargs.get('with_header', None)
        if with_header is None:
            with_header = snif.has_header(stdout)
        if with_header:
            hds = next(rd) if stdout else ''

            xlog.debug("+++++ {0}".format(hds))
            if (len(hds) > 0) and (hds[0].find(
                    'CMMVC7017E Login has failed') == 0):
                error_msg = ('CLI failure. Return code is 1. '
                             'Error message is "{0}"'.format(hds[0]))
                xlog.error(error_msg)
                xlog.debug("+++++ STDOUT: \n{0}".format(stdout))
                xlog.debug("+++++ STDERR: \n{0}".format(stderr))
                raise CLIFailureError(error_msg, returnCode=1)

            for row in rd:
                cur = Munch()
                for k, v in zip(hds, row):
                    append_dict(cur, k, v, strip=True)
                result.append(cur)
        else:
            cur = Munch()
            for row in rd:
                if row:
                    append_dict(cur, row[0], ' '.join(row[1:]), strip=True)
                elif cur:  # start new section
                    result.append(cur)
                    cur = Munch()
            if cur:
                result.append(cur)
        return result

    def __iter__(self):
        return iter(self.result)

    @property
    def as_single_element(self):
        '''Return response to a single dict like object
           :py:class:`xiv.dtypes.Bunch` or None.'''
        return reduce(
            merge_dict,
            self.result,
            Munch()) if self.result else None

    @property
    def as_list(self):
        '''Return response as a list of :py:class:`xiv.dtypes.Bunch`.'''
        return self.result[:]

    def as_dict(self, key):
        '''Return response which has the key to a dict like
           object :py:class:`xiv.dtypes.Bunch`.'''
        res = Munch()
        _ = [append_dict(res, a[key], a) for a in self.result if key in a]
        return compact_dict(res)


class SVCResponse(CLIResponse):
    resp_type = 'svc'


class svcinfo_lsroute_response(SVCResponse):
    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        if not kwargs.get('delim', None):
            kwargs['delim'] = ' '

        def sections():
            cur = []
            for line in stdout.splitlines():
                if 'routing table' in line:
                    if cur:
                        cur[0:1] = [cur[0].replace(' Next Hop ', ' Next_Hop ')]
                        yield ('\n'.join(cur), stderr)
                        cur = []
                else:
                    line = line.strip()
                    if line:
                        cur.append(line)
            if cur:
                cur[0:1] = [cur[0].replace(' Next Hop ', ' Next_Hop ')]
                yield ('\n'.join(cur), stderr)
        return [d for sec in sections()
                for d in SVCResponse.parse(self, sec, kwargs)]

# response helper to parse vvolmd_entry_create, vvolmd_entry_update,
# vvolmd_entry_retrieve command#


class svctask_metadata_entry_response(SVCResponse):
    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        stdout = stdout.lstrip()  # remove starting space to work with csv
        stdoutlines = stdout.splitlines()

        # e.g. if stdout is "<error_tag> 4", the return code is 4
        error_tag = kwargs.get('error_tag', '')
        if error_tag:
            for line in stdoutlines:
                idx = line.find(error_tag)
                if idx >= 0:
                    rc = ' '.join(tk for tk in
                                  line[idx + len(error_tag):].split() if
                                  tk.isdigit() or tk.startswith('-')
                                  and tk[1:].isdigit())
                    raise CLIFailureError(
                        'CLI failure. Return code is %s. Error message '
                        'is "%s"' % (rc, stderr), returnCode=int(rc))

        # start to deal with raw response
        result = []
        # entry create, update
        if len(stdoutlines) == 1:
            result.append(colon2Bunch(stdoutlines[0]))
        # entry retrieve
        elif len(stdoutlines) > 1:
            entry = colon2Bunch(stdoutlines[0])
            stdoutlines.pop(0)
            idx = stdout.find('\n')
            entry['content'] = stdout[idx + 1:]
            result.append(entry)
        return result


# response helper to parse vvolmd_entry_list command#
class svctask_metadata_entry_list_response(SVCResponse):
    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        stdout = stdout.lstrip()  # remove starting space to work with csv
        stdoutlines = stdout.splitlines()

        # e.g. if stdout is "<error_tag> 4", the return code is 4
        error_tag = kwargs.get('error_tag', '')
        if error_tag:
            for line in stdoutlines:
                idx = line.find(error_tag)
                if idx >= 0:
                    rc = ' '.join(tk for tk in
                                  line[idx + len(error_tag):].split()
                                  if tk.isdigit() or tk.startswith('-')
                                  and tk[1:].isdigit())
                    raise CLIFailureError(
                        'CLI failure. Return code is %s. Error message '
                        'is "%s"' % (rc, stderr), returnCode=int(rc))

        # start to deal with raw response
        result = []
        try:
            if len(stdoutlines) < 1:
                return result
            label = stdoutlines.pop(0)
            label = label.split()
            for line in stdoutlines:
                entry = Munch()
                line_entry = line.split()
                for i in [0, 1, 2]:
                    entry[label[i]] = line_entry[i]
                result.append(entry)
        except Exception:
            raise CLIFailureError('Fail to parse the CLI response :' + stdout)
            return result
        return result


# response helper to parse lsmetadatavdisk command#
class svcinfo_lsmetadatavdisk_response(SVCResponse):
    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        stdout = stdout.lstrip()  # remove starting space to work with csv
        stdoutlines = stdout.splitlines()

        # e.g. if stdout is "<error_tag> 4", the return code is 4
        error_tag = kwargs.get('error_tag', '')
        if error_tag:
            for line in stdoutlines:
                idx = line.find(error_tag)
                if idx >= 0:
                    rc = ' '.join(tk for tk in
                                  line[idx + len(error_tag):].split() if
                                  tk.isdigit() or tk.startswith('-')
                                  and tk[1:].isdigit())
                    raise CLIFailureError(
                        'CLI failure. Return code is %s. Error message '
                        'is "%s"' % (rc, stderr), returnCode=int(rc))

        # start to deal with raw response
        result = []
        result.append(lines2Bunch(stdoutlines))
        return result

# response helper to parse vvolmd_db_list command


class svctask_metadata_db_list_response(SVCResponse):
    def parse(self, resp, kwargs):
        if isinstance(resp, basestring):
            stdout, stderr = resp, ''
        else:
            stdout, stderr = resp
        if isinstance(stdout, bytes):
            stdout = stdout.decode()
        if isinstance(stderr, bytes):
            stderr = stderr.decode()
        stdout = stdout.lstrip()  # remove starting space to work with csv
        stdoutlines = stdout.splitlines()

        # e.g. if stdout is "<error_tag> 4", the return code is 4
        error_tag = kwargs.get('error_tag', '')
        if error_tag:
            for line in stdoutlines:
                idx = line.find(error_tag)
                if idx >= 0:
                    rc = ' '.join(tk for tk in
                                  line[idx + len(error_tag):].split() if
                                  tk.isdigit() or tk.startswith('-') and
                                  tk[1:].isdigit())
                    raise CLIFailureError(
                        'CLI failure. Return code is %s. Error message '
                        'is "%s"' % (rc, stderr), returnCode=int(rc))

        result = []
        try:
            if len(stdoutlines) < 1:
                return result
            label = stdoutlines.pop(0)
            label = label.split()
            for line in stdoutlines:
                entry = Munch()
                line_entry = line.split()
                for i in [0]:
                    entry[label[i]] = line_entry[i]
                result.append(entry)
        except Exception:
            raise CLIFailureError('Fail to parse the CLI response :' + stdout)
            return result
        return result


class MySniffer(csv.Sniffer):
    def __init__(self, delim):
        csv.Sniffer.__init__(self)
        self.delim = delim

    def sniff(self, sample, delimiters=None):
        class dialect(csv.Dialect):
            delimiter = self.delim
            quotechar = '"'
            doublequote = True
            skipinitialspace = True
            lineterminator = '\r\n'
            quoting = csv.QUOTE_MINIMAL
        return dialect

    def has_header(self, sample):
        ss = StringIO(sample)
        try:
            return self._has_header(csv.reader(ss, self.sniff(sample)))
        except Exception:
            xlog.exception('Can not detect header, and continue.')
        finally:
            ss.close()
        return False

    def _has_header(self, rdr):
        try:
            header = next(rdr)
        except StopIteration:
            return False
        columns = len(header)
        coltypes = {}
        rows = [row for row in rdr if len(row) == columns][:20]
        for col in range(columns):
            if all(is_numeric(row[col]) for row in rows):
                coltypes[col] = int
            elif len(set(len(row[col]) for row in rows)
                     ) == 1:  # all cells in column have same length
                coltypes[col] = len(rows[0][col])
        hasHeader = 0
        for col, tp in list(coltypes.items()):
            if tp is int:
                hasHeader += -1 if is_numeric(header[col]) else 1
            else:
                hasHeader += -1 if len(header[col]) == tp else 1
        return hasHeader > 0


def compare_similar(data1, data2, factor=2):
    '''Return True if most of keys in both data are same, otherwise False'''
    la, lb = len(data1), len(data2)
    lc = len(set(data1.keys()).intersection(set(data2.keys()))) * factor
    return lc >= max(la, lb)


def append_dict(dict_, key, value, strip=False):
    if strip:
        key, value = key.strip() if isinstance(
            key, basestring) else key, value.strip() if isinstance(
            value, basestring) else value
    obj = dict_.get(key, None)
    if obj is None:
        dict_[key] = value
    elif isinstance(obj, list):
        obj.append(value)
        dict_[key] = obj
    else:
        dict_[key] = [obj, value]
    return dict_


def compact_dict(dict_):
    for k, v in list(dict_.items()):
        if isinstance(v, list) and v:
            for a, b in zip(v, v[1:]):
                if a != b:
                    break
            else:  # if all members are same
                dict_[k] = v[0]
    return dict_


def merge_dict(dict1, dict2):
    for k, v in list(dict2.items()):
        append_dict(dict1, k, v)
    return compact_dict(dict1)


def compact_if_not_similar(dicts):
    for a, b in zip(dicts, dicts[1:]):
        # if some data are not similar, they are aspects of the same object,
        # so merge all data
        # e.g. "svcinfo lsvdisk 1"
        if not compare_similar(a, b):
            return [reduce(merge_dict, dicts, Munch())]
    # if all data are similar, they are for different objects, so no merge
    # e.g. "svcinfo lsportip 1"
    return dicts


def is_numeric(data):
    for t in (to_long_hex, float):
        try:
            t(data)
            return True
        except (ValueError, OverflowError):
            pass
    return False

# " Token: 1234 TimeStamp: 20141013" =>
# Bunch(Token='1234',TimeStamp='20141013')


def colon2Bunch(line=''):
    ret = Munch()
    pairs = line.split()
    i = 0
    while i < len(pairs):
        if pairs[i].endswith(':'):
            ret[pairs[i].replace(':', '')] = pairs[i + 1]
            i = i + 2
    return ret

# ["accesspath /mnt/file","vdisk_id    1"] =>
# Bunch(accesspath='/mnt/file',vdisk_id='1')


def lines2Bunch(lines=[]):
    ret = Munch()
    for line in lines:
        pairs = line.split()
        i = 0
        while i < len(pairs):
            ret[pairs[i]] = pairs[i + 1]
            i = i + 2
    return ret


def to_long_hex(data):
    return long(data, 16)


def is_svc_response(name):
    return getattr(name, 'resp_type', None) == 'svc' or (
        isinstance(name, basestring) and name.startswith('svc'))


def find_response_helper(name, param=None):
    '''Find CLI response helper.

    :param name: The name of CLI response type. It has "svc" prefix for
                 SVC CLI.
    :type name: str
    :param param: (optional) Extra parameter.
    :type param: str
    '''
    if is_svc_response(name):
        if param in [
            'metadata_entry_create',
            'metadata_entry_update',
                'metadata_entry_retrieve']:
            return svctask_metadata_entry_response
        elif param == 'metadata_entry_list':
            return svctask_metadata_entry_list_response
        elif param == 'metadata_db_list':
            return svctask_metadata_db_list_response
        elif param == 'lsmetadatavdisk':
            return svcinfo_lsmetadatavdisk_response
        elif param == 'lsroute':
            return svcinfo_lsroute_response
        else:
            return SVCResponse



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import xml.etree.ElementTree as et
import xml.etree.cElementTree as cet
from contextlib import contextmanager
from xml.parsers.expat import ExpatError

try:
    basestring
except NameError:
    basestring = str


Element = cet.Element
tostring = cet.tostring


class XMLException(Exception):
    pass


class ElementNotFoundException(XMLException):

    def __init__(self, xml, notFound):
        XMLException.__init__(self)
        self.xml = xml
        self.notFound = notFound

    def __str__(self):
        return "Cannot parse XML (cannot find %s):\n%s" % (
            self.notFound, tostring(self.xml))


def str_brief(obj, lim=20, dots='...', use_repr=True):
    """Truncates a string, starting from 'lim' chars. The given object
    can be a string, or something that can be casted to a string.

    >>> import string
    >>> str_brief(string.uppercase)
    'ABCDEFGHIJKLMNOPQRST...'
    >>> str_brief(2 ** 50, lim=10, dots='0')
    '11258999060'
    """
    if isinstance(obj, basestring) or not use_repr:
        full = str(obj)
    else:
        full = repr(obj)
    postfix = []
    CLOSERS = {'(': ')', '{': '}', '[': ']', '"': '"', "'": "'", '<': '>'}
    for i, c in enumerate(full):
        if i >= lim + len(postfix):
            return full[:i] + dots + ''.join(reversed(postfix))
        if postfix and postfix[-1] == c:
            postfix.pop(-1)
            continue
        closer = CLOSERS.get(c, None)
        if closer is not None:
            postfix.append(closer)
    return full


class XMLSyntaxError(XMLException):

    """
    A somewhat friendlier XML Syntax Error.

    >>> fromstring('<a/>').tostring()
    '<a/>'
    >>> tag='<tag description="a very long description that will be briefed">'
    >>> fromstring(tag) # doctest: +ELLIPSIS
    Traceback (most recent call last):
       ...
    XMLSyntaxError: Malformed XML (line 1, XML '<tag description...>'): ...
    >>> fromstring('')
    Traceback (most recent call last):
       ...
    XMLSyntaxError: Malformed XML (line None, XML ''): No specific message
    >>>
    """

    def __init__(self, original, msg, lineno):
        XMLException.__init__(self)
        self.original = original
        self.msg = msg or "No specific message"
        self.lineno = lineno

    def pretty(self):
        return ("Syntax error parsing XML:\n\tLine#: %s\n\tMessage: %s\n\t"
                "XML:%r" % (self.lineno, self.msg, self.original))

    def __str__(self):
        return "Malformed XML (line %s, XML %r): %s" % (
            self.lineno, str_brief(self.original, lim=15), self.msg)


@contextmanager
def _translateExceptions(original):
    try:
        yield None
    except ExpatError as e:
        raise XMLSyntaxError(original, e.args[0], e.lineno)
    except (cet.ParseError, et.ParseError) as e:
        raise XMLSyntaxError(original, e.args[0], e.lineno)


def fromstring(text):
    with _translateExceptions(None):
        return cet.fromstring(text)


def parse(obj):
    with _translateExceptions(None):
        return cet.parse(obj)


def xml_find(elem, path, attrib=None):
    elem2 = elem.find(path)
    if elem2 is None:
        raise ElementNotFoundException(elem, path)
    if attrib is None:
        return elem2
    else:
        if attrib not in elem2.attrib:
            raise ElementNotFoundException(elem, "%s/@%s" % (path, attrib))
        return elem2.attrib[attrib]

# =========================================================================
# TerminationDetectingXMLParser
# =========================================================================


class _TerminationDetectingTreeBuilder(et.TreeBuilder):

    def __init__(self):
        et.TreeBuilder.__init__(self)
        self.root_element = None
        self.root_element_closed = False

    def start(self, tag, attrs):
        element = et.TreeBuilder.start(self, tag, attrs)
        if self.root_element is None:
            self.root_element = element
        return element

    def end(self, tag):
        element = et.TreeBuilder.end(self, tag)
        if self.root_element is element:
            self.root_element_closed = True
        return element


class TerminationDetectingXMLParser(object):

    """An XML parser which you can feed from a stream; knows automatically
    when the first tag was closed"

    >>> td = TerminationDetectingXMLParser()
    >>> td.feed('<a>')
    >>> td.root_element_closed
    False
    >>> td.feed('</a>')
    >>> td.root_element_closed
    True
    >>> td.close().tostring()
    '<a/>'

    >>> td = TerminationDetectingXMLParser()
    >>> td.feed('<a><b></b></a>')
    >>> td.root_element_closed
    True
    >>> td.close().tostring()
    '<a><b/></a>'

    >>> td = TerminationDetectingXMLParser()
    >>> td.feed('<a<a') #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    ParseError: not well-formed (invalid token): line 1, column 2
    >>>
    """

    def __init__(self):
        self.tree_builder = _TerminationDetectingTreeBuilder()
        self.xml_tree_builder = et.XMLParser(target=self.tree_builder)

    def feed(self, chunk):
        with _translateExceptions(chunk):
            self.xml_tree_builder.feed(chunk)

    def close(self):
        with _translateExceptions(None):
            tree = self.xml_tree_builder.close()
            return fromstring(tostring(tree))

    @property
    def root_element_closed(self):
        return self.tree_builder.root_element_closed



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

# ------------- Transport ---------------------


class TransportMessages(object):
    TRANSPORT_DISCONNECT = 'Transport is disconnected'
    SSH_INCORRECT_PRIVATE_KEY = 'Private key incorrect'
    SSH_FAILED_TO_LOAD_KNOWN_HOSTS = 'Failed to load known hosts file'
    SSH_PASS_PHRASE_REQUIRED = 'Check if password was specified to unlock ' \
                               'the private key file'
    SSH_AUTHENTICATION_FAILURE = 'Authentication failed'
    SSH_CONNECT_TIMED_OUT = 'Timed out when attempting to establish ' \
                            'a ssh connection'
    SSH_CON_TIMED_OUT_WHEN_EXEC_CMD = 'Timed out when executing command ' \
                                      'through a ssh connection, possibly ' \
                                      'the command is interactive'

    @staticmethod
    def SSH_UNABLE_TO_CONNECT(host):
        return 'Cannot establish ssh connection to %s' % host

    @staticmethod
    def SSH_ALLOWED_AUTH_TYPES(types):
        return 'Allowed authentication types are: %s' % types

    @staticmethod
    def SSH_BAD_HOST_FINGERPRINT(host):
        return 'Host finger print does not match when connect to %s. Add ' \
               'correct host key in xsf_known_hosts to get rid of ' \
               'this exception.' % host

    @staticmethod
    def SSH_HOST_DOES_NOT_EXIST(host):
        return 'Cannot establish the ssh connection to %s due to ' \
               'unrecognizable host' % host

# ------------- Unified ---------------------


class UnifiedMessages(object):
    UNIFIED_PARSE_LOCAL_START = 'Try to parse local CLI specification'
    UNIFIED_NO_CLI_SPEC = 'No CLI specification is not found'
    UNIFIED_PARSE_REMOTE_FAIL = 'It fails to parse the CLI specification ' \
                                'from storage array.'
    SSH_NO_ENDPOINT_PROVIDED = 'No ssh connection endpoint provided'
    SVCCLI_UNKNOWN_ARRAY = 'SVCCLI: Unknown array'

    @staticmethod
    def UNIFIED_INCORRECT_DEVICE_TYPE(device):
        return 'The expected device %s is not found.' % (device)

    @staticmethod
    def UNIFIED_CATXMLSPEC_FAIL(err):
        return 'The storage array fails to run "catxmlspec" command: %s' % (
            err)



"""
Usage examples of the SVC CLI Client.
"""

from pysvc.unified.client import connect

# Connect storage array through SSH
# It supports two authentication methods
# 1) use SSH prviate key to login
#    Prerequisite:
#     configure SSH Public Key to the storage user (e.g. admin)
#     from storage GUI(Access --> Users --> Create user with the SSH Pubic Key)
svccli = connect('ip', username='admin', privatekey_filename=r'/local/key')

# 2) use username and password to login
svccli = connect('ip_address', username='username', password='password')

# Create volume
svccli.svctask.mkvolume(name="vol_name", pool="pool", size=1, unit='gb')

# Get Volume from storage
vol = svccli.svcinfo.lsvdisk(object_id='vol_name').as_single_element
print(vol.name)

# Get all the volumes from storage as a list element
vols = svccli.svcinfo.lsvdisk(bytes=True).as_list
for vol in vols:
    print(vol.name)

# Get Pool from storage
pool = svccli.svcinfo.lsmdiskgrp(object_id='pool_name').as_single_element
print(pool.name)

# Getting all of the pools from storage as a list element
pools = svccli.svcinfo.lsmdiskgrp().as_list
for p in pools:
    print(p.name)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
'''Unified SSH client for IBM Spectrum Virtualize Family Storage

Example:

>>> from pysvc.unified import connect
>>> conn = connect('ip', username='admin', privatekey_filename=r'/local/key')
>>> conn.get_device_info() # get device information of storage array
('svc', [('svc', '6.2')])
>>>
>>> dir(conn) # list available commands
[..., 'svcinfo', ...]
>>> dir(conn.svcinfo)
[..., 'lscluster', ...]
>>>
>>> print conn.svcinfo.lscluster.__doc__ # get help
Help on lscluster(**kwargs): ...
>>>
>>> for cluster in conn.svcinfo.lscluster():
...     print cluster
...
..., name = 'cluster1', ...
..., name = 'cluster2', ...
..., name = 'cluster3', ...
>>>
>>> conn.close() # close the connection
'''

from pysvc.unified.client import connect

__all__ = ['connect']



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pysvc.messages import TransportMessages


class CommonTransport(object):
    def connect(self):
        raise NotImplementedError(
            "This interface should be called from subclass instances.")

    def send_command(self, command):
        raise NotImplementedError(
            "This interface should be called from subclass instances.")

    def __repr__(self):
        if self.connected_endpoint:
            return '<%s (%s)>' % (self.__class__.__name__,
                                  self.connected_endpoint)
        return '<%s>' % self.__class__.__name__

    def __str__(self):
        if self.connected_endpoint:
            return str(self.connected_endpoint)
        return TransportMessages.TRANSPORT_DISCONNECT



from socket import timeout as _SocketTimeout
from contextlib import closing
from pysvc.unified.errors import SCPError, SCPTimeoutError

MSG_PART = 3
SIZE_INDEX = 1
CMD_INDEX = 0
SCP_CMD = 'scp -f {path}'
MSG_BUF_DEFUALT_SIZE = 16*1024
FILE_BUF_DEFUALT_SIZE = 64*1024


class ScpClient(object):

    def __init__(self, transport, timeout=None):
        """
        :param transport: ssh transport, e.g.
                client = paramiko.SSHClient()
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                client.connect(hostname, username=username, password=password)
                self.transport = client.get_transport()
        :param timeout: int, timeout for ssh session
        :return:
        """
        self.transport = transport
        self.timeout = timeout
        # The buf for msg, should greater than msg size
        self.msg_buf_size = MSG_BUF_DEFUALT_SIZE
        # The buf for file, should greater then file size
        self.file_buf_size = FILE_BUF_DEFUALT_SIZE

    def set_msg_buf_size(self, size):
        self.msg_buf_size = size

    def set_file_buf_size(self, size):
        self.file_buf_size = size

    def receive(self, remote_path):
        """
        scp the remote_path
        :param remote_path: str, the full path of the remote file
        :return: str, context of the remote file
        """

        result = None
        scp_cmd = SCP_CMD.format(path=remote_path)
        with closing(self._ssh_open_channel()) as channel:
            # Execute the scp command on the far side.
            channel.exec_command(scp_cmd)
            while not channel.closed:
                channel.sendall('\x00')
                msg = self._scp_recv(channel, self.msg_buf_size)

                if not msg:
                    break

                cmd = msg[CMD_INDEX]
                msg = msg[SIZE_INDEX:]

                if cmd == 'T':
                    pass
                elif cmd == 'C':
                    # Receive file.
                    result = self._scp_receive_file(channel, msg)
                    break
                elif cmd == '\x01':
                    raise SCPError('scp error: {0!r}'.format(msg))
                else:
                    raise SCPError('Unknown scp reply: {0!r} {1!r}'
                                   .format(cmd, msg))

            return result

    def _ssh_open_channel(self):
        """
        Open the ssh session for SCP
        :return: ssh session
        """
        channel = self.transport.open_session()
        if self.timeout is not None:
            channel.settimeout(self.timeout)
        return channel

    @staticmethod
    def _scp_recv(channel, buf_size):
        """
        Receive a part (buf_size byte) of file
        :param buf_size: int, the size of buffer to receive
        :return: str, the receive context
        """
        try:
            return channel.recv(buf_size).decode()
        except _SocketTimeout:
            raise SCPTimeoutError('Timeout waiting for scp response')

    def _scp_receive_file(self, channel, msg):
        """
        Read a remote file, yield parts of the file, until EOF.
        :param channel: ssh session obj,
        :param msg: str, the status of file
        :return: str, the context of the file
        """

        size = self._extract_file_size_info(msg)
        result = ""
        try:
            # Tell the remote side we're ready to read
            channel.sendall('\x00')

            bytes_read = 0
            while bytes_read < size:
                # Compute the max to read.
                bytes_to_read = self.file_buf_size
                if size - bytes_read <= self.file_buf_size:
                    bytes_to_read = size - bytes_read

                s = self._scp_recv(channel, bytes_to_read)

                result += s

                bytes_read += len(s)

            # determine the end of the file
            msg = self._scp_recv(channel, self.msg_buf_size)
            if len(msg) == 0:
                raise SCPError('Error on end of read: msg is empty')

            if msg[CMD_INDEX] != '\x00':
                raise SCPError('Error on end of read: {0!r} {1!r}'.format(
                    msg[CMD_INDEX], msg[SIZE_INDEX:]))

        except _SocketTimeout:
            raise SCPTimeoutError('Timeout on file read')
        return result

    @staticmethod
    def _extract_file_size_info(msg):
        """
        :param msg: str, should contain 'mode size pathname'
        :return: int, the size in the msg
        """

        parts = msg.split()

        if len(parts) != MSG_PART:
            raise SCPError('Invalid file receive header: {0!r}'.format(msg))
        try:
            size = int(parts[SIZE_INDEX])
        except ValueError:
            raise SCPError('Bad file size: {0!r}'.format(parts[1]))

        return size



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
from pysvc.messages import TransportMessages


class StorageArrayClientException(Exception):
    """
    Base class for generic storge array cleint exceptions.
    """

    def __init__(self, message=None, original_exception=None):
        if message is None and original_exception is None:
            raise ValueError(
                "either message or original_exception must be set")
        self.my_message = message
        self.original_exception = original_exception

    def __str__(self):
        response = ""
        if self.my_message:
            response += str(self.my_message)
        if self.original_exception is not None:
            response += "::" + "::".join(str(ex)
                                         for ex in self.original_exception)

        return response

    def __unicode__(self):
        return str(self.__str__())

# --------------------  Authentication  --------------------


class IncorrectCredentials(StorageArrayClientException):
    pass


class IncorrectUsername(IncorrectCredentials):
    pass


class IncorrectPassword(IncorrectCredentials):
    pass


class UnsupportXmlCompression(StorageArrayClientException):
    pass

# --------------------  SSH Authentication  --------------------


class SSHFailedToLoadKnownHosts(IncorrectCredentials):
    pass


class SSHAuthenticationFailure(IncorrectCredentials):
    pass


class BadAuthenticationTypeException(SSHAuthenticationFailure):
    """
    Exception raised when incorrect authentication types were used.
    """

    def __init__(
            self,
            message=None,
            allowed_types=[],
            original_exception=None):
        super(
            BadAuthenticationTypeException,
            self).__init__(
            message,
            original_exception)
        self._allowed_auth_types = allowed_types

    @property
    def allowed_authentication_types(self):
        return self._allowed_auth_types

    def __str__(self):
        return TransportMessages.SSH_ALLOWED_AUTH_TYPES(
            self._allowed_auth_types)


class PassphraseRequiredException(SSHAuthenticationFailure):
    """
    Exception raised when a pass phrase is absent when
    attempting to unlock the private key.
    """
    pass


class PartialAuthenticationException(SSHAuthenticationFailure):
    """
    Exception raised when a partial authentication was detected.
    """

    def __init__(
            self,
            message=None,
            allowed_types=[],
            original_exception=None):
        super(
            PartialAuthenticationException,
            self).__init__(
            message,
            original_exception)
        self._allowed_auth_types = allowed_types

    @property
    def allowed_authentication_types(self):
        return self._allowed_auth_types

    def __str__(self):
        return TransportMessages.SSH_ALLOWED_AUTH_TYPES(
            self._allowed_auth_types)


class BadHostFingerPrintException(SSHAuthenticationFailure):
    """
    Host finger print given by the server did not match expected one.
    """

    def __init__(
            self,
            message=None,
            hostname=None,
            expected_key=None,
            presented_key=None,
            original_exception=None):
        super(
            BadHostFingerPrintException,
            self).__init__(
            message,
            original_exception)
        self._hostname = hostname
        self._expeted_key = expected_key
        self._presented_key = presented_key
        self._original_exception = original_exception

    @property
    def presented_key(self):
        return self._presented_key

    @property
    def hostname(self):
        return self._hostname

    @property
    def expeted_key(self):
        return self._expected_key

    def __str__(self):
        return TransportMessages.SSH_BAD_HOST_FINGERPRINT(self._hostname)

# --------------------  Connection Error --------------------


class UnableToConnectException(StorageArrayClientException):
    pass


class HostDoesNotExistException(UnableToConnectException):
    """
    Raised when the destination host does not exist.
    """

    def __init__(self, message=None, hostname=None, original_exception=None):
        super(
            HostDoesNotExistException,
            self).__init__(
            message,
            original_exception)
        self._hostname = hostname

    @property
    def hostname(self):
        return self._hostname

    def __str__(self):
        return TransportMessages.SSH_HOST_DOES_NOT_EXIST(self._hostname)


class ConnectionTimedoutException(UnableToConnectException):
    pass


class ProtocolMismatchException(UnableToConnectException):
    pass


class CantInstantiateError(StorageArrayClientException):
    pass

# --------------------  SSL Authentication  --------------------


class SSLAuthenticationFailure(IncorrectCredentials):
    """
    TODO: Combine SSL exceptions into this hierarchy.
    """
    pass

# --------------------  Execution --------------------


class GeneralExecutionError(Exception):
    """
    Base class for generic storge array call execution exception.
    """

    def __init__(self, message=None, original_exception=None):
        if message is None and original_exception is None:
            raise ValueError(
                "either message or original_exception must be set")
        self.my_message = message
        self.original_exception = original_exception

    def _get_error_string(self, error):
        response = ""

        if not isinstance(
                error,
                GeneralExecutionError) or isinstance(
                error,
                CommandExecutionError):
            response += str(error)
        elif hasattr(error, 'my_message'):
            response += str(error.my_message)
        if hasattr(
                error,
                'original_exception') and error.original_exception is not None:
            response += "::" + "::".join(self._get_error_string(ex)
                                         for ex in error.original_exception)

        return response

    def __str__(self):
        return self._get_error_string(self)

    def __unicode__(self):
        return str(self._get_error_string(self))


class UnknownDeviceTypeError(GeneralExecutionError):
    pass


class CommandExecutionError(GeneralExecutionError):
    """
    Base class for storge array command execution exception.
    Do NOT call __str__ from _get_error_string in this class
    and its sub-classes.
    """

    def __init__(
            self,
            message=None,
            command=None,
            response=None,
            original_exception=None):
        super(
            CommandExecutionError,
            self).__init__(
            message,
            original_exception)
        self.my_command = command
        self.my_response = response

    def _get_error_string(self, error):
        response = ""
        if hasattr(error, 'my_command'):
            response += "Storage array command <" + \
                str(error.my_command) + "> failed: "
        if hasattr(error, 'my_response'):
            response += str(error.my_response)

        return response


class ResponseParserError(CommandExecutionError):
    pass


class InternalSystemError(CommandExecutionError):
    pass
